@@ -21,6 +21,7 @@
 #include "packer.h"
 #include "packer_class.h"
 #include "buffer_class.h"
+#include "factory_class.h"
 
 VALUE cMessagePack_Packer;
 
@@ -42,24 +43,32 @@ static void Packer_free(msgpack_packer_t* pk)
     if(pk == NULL) {
         return;
     }
+    msgpack_packer_ext_registry_destroy(&pk->ext_registry);
     msgpack_packer_destroy(pk);
-    free(pk);
+    xfree(pk);
 }
 
-static VALUE Packer_alloc(VALUE klass)
+static void Packer_mark(msgpack_packer_t* pk)
+{
+    msgpack_packer_mark(pk);
+    msgpack_packer_ext_registry_mark(&pk->ext_registry);
+}
+
+VALUE MessagePack_Packer_alloc(VALUE klass)
 {
     msgpack_packer_t* pk = ALLOC_N(msgpack_packer_t, 1);
     msgpack_packer_init(pk);
 
-    VALUE self = Data_Wrap_Struct(klass, msgpack_packer_mark, Packer_free, pk);
+    VALUE self = Data_Wrap_Struct(klass, Packer_mark, Packer_free, pk);
 
     msgpack_packer_set_to_msgpack_method(pk, s_to_msgpack, self);
+    msgpack_packer_ext_registry_init(&pk->ext_registry);
     pk->buffer_ref = MessagePack_Buffer_wrap(PACKER_BUFFER_(pk), self);
 
     return self;
 }
 
-static VALUE Packer_initialize(int argc, VALUE* argv, VALUE self)
+VALUE MessagePack_Packer_initialize(int argc, VALUE* argv, VALUE self)
 {
     VALUE io = Qnil;
     VALUE options = Qnil;
@@ -88,14 +97,7 @@ static VALUE Packer_initialize(int argc, VALUE* argv, VALUE self)
 
     PACKER(self, pk);
 
-    MessagePack_Packer_initialize(pk, io, options);
-
-    return self;
-}
-
-void MessagePack_Packer_initialize(msgpack_packer_t* pk, VALUE io, VALUE options)
-{
-    MessagePack_Buffer_initialize(PACKER_BUFFER_(pk), io, options);
+    MessagePack_Buffer_set_options(PACKER_BUFFER_(pk), io, options);
 
     if(options != Qnil) {
         VALUE v;
@@ -103,6 +105,14 @@ void MessagePack_Packer_initialize(msgpack_packer_t* pk, VALUE io, VALUE options
         v = rb_hash_aref(options, ID2SYM(rb_intern("compatibility_mode")));
         msgpack_packer_set_compat(pk, RTEST(v));
     }
+
+    return self;
+}
+
+static VALUE Packer_compatibility_mode_p(VALUE self)
+{
+    PACKER(self, pk);
+    return pk->compatibility_mode ? Qtrue : Qfalse;
 }
 
 static VALUE Packer_buffer(VALUE self)
@@ -139,6 +149,29 @@ static VALUE Packer_write_map_header(VALUE self, VALUE n)
     return self;
 }
 
+static VALUE Packer_write_float32(VALUE self, VALUE numeric)
+{
+    if(!rb_obj_is_kind_of(numeric, rb_cNumeric)) {
+      rb_raise(rb_eArgError, "Expected numeric");
+    }
+
+    PACKER(self, pk);
+    msgpack_packer_write_float(pk, (float)rb_num2dbl(numeric));
+    return self;
+}
+
+static VALUE Packer_write_ext(VALUE self, VALUE type, VALUE data)
+{
+    PACKER(self, pk);
+    int ext_type = NUM2INT(type);
+    if(ext_type < -128 || ext_type > 127) {
+        rb_raise(rb_eRangeError, "integer %d too big to convert to `signed char'", ext_type);
+    }
+    StringValue(data);
+    msgpack_packer_write_ext(pk, ext_type, data);
+    return self;
+}
+
 static VALUE Packer_flush(VALUE self)
 {
     PACKER(self, pk);
@@ -201,51 +234,81 @@ static VALUE Packer_write_to(VALUE self, VALUE io)
 //    return self;
 //}
 
+static VALUE Packer_registered_types_internal(VALUE self)
+{
+    PACKER(self, pk);
+#ifdef HAVE_RB_HASH_DUP
+    return rb_hash_dup(pk->ext_registry.hash);
+#else
+    return rb_funcall(pk->ext_registry.hash, rb_intern("dup"), 0);
+#endif
+}
+
+static VALUE Packer_register_type(int argc, VALUE* argv, VALUE self)
+{
+    PACKER(self, pk);
+
+    int ext_type;
+    VALUE ext_class;
+    VALUE proc;
+    VALUE arg;
+
+    switch (argc) {
+    case 2:
+        /* register_type(0x7f, Time) {|obj| block... } */
+        rb_need_block();
+#ifdef HAVE_RB_BLOCK_LAMBDA
+        proc = rb_block_lambda();
+#else
+        /* MRI 1.8 */
+        proc = rb_block_proc();
+#endif
+        arg = proc;
+        break;
+    case 3:
+        /* register_type(0x7f, Time, :to_msgpack_ext) */
+        arg = argv[2];
+        proc = rb_funcall(arg, rb_intern("to_proc"), 0);
+        break;
+    default:
+        rb_raise(rb_eArgError, "wrong number of arguments (%d for 2..3)", argc);
+    }
+
+    ext_type = NUM2INT(argv[0]);
+    if(ext_type < -128 || ext_type > 127) {
+        rb_raise(rb_eRangeError, "integer %d too big to convert to `signed char'", ext_type);
+    }
+
+    ext_class = argv[1];
+    if(rb_type(ext_class) != T_CLASS) {
+        rb_raise(rb_eArgError, "expected Class but found %s.", rb_obj_classname(ext_class));
+    }
+
+    msgpack_packer_ext_registry_put(&pk->ext_registry, ext_class, ext_type, proc, arg);
+
+    if (ext_class == rb_cSymbol) {
+        pk->has_symbol_ext_type = true;
+    }
+
+    return Qnil;
+}
+
 VALUE MessagePack_pack(int argc, VALUE* argv)
 {
     VALUE v;
-    VALUE io = Qnil;
-    VALUE options = Qnil;
 
-    if(argc == 1) {
-        v = argv[0];
-
-    } else if(argc == 2) {
-        v = argv[0];
-        if(rb_type(argv[1]) == T_HASH) {
-            options = argv[1];
-        } else {
-            io = argv[1];
-        }
-
-    } else if(argc == 3) {
-        v = argv[0];
-        io = argv[1];
-        options = argv[2];
-        if(rb_type(options) != T_HASH) {
-            rb_raise(rb_eArgError, "expected Hash but found %s.", rb_obj_classname(options));
-        }
-
-    } else {
+    if (argc < 0 || argc > 3) {
         rb_raise(rb_eArgError, "wrong number of arguments (%d for 1..3)", argc);
     }
+    v = argv[0];
 
-    VALUE self = Packer_alloc(cMessagePack_Packer);
+    VALUE self = MessagePack_Factory_packer(argc - 1, argv + 1, cMessagePack_DefaultFactory);
     PACKER(self, pk);
 
-    if (options != Qnil) {
-      pk->compatibility_mode = RTEST(rb_hash_aref(options, ID2SYM(rb_intern("compatibility_mode"))));
-    }
-
-    //msgpack_packer_reset(s_packer);
-    //msgpack_buffer_reset_io(PACKER_BUFFER_(s_packer));
-
-    MessagePack_Packer_initialize(pk, io, options);
-
     msgpack_packer_write_value(pk, v);
 
     VALUE retval;
-    if(io != Qnil) {
+    if(msgpack_buffer_has_io(PACKER_BUFFER_(pk))) {
         msgpack_buffer_flush(PACKER_BUFFER_(pk));
         retval = Qnil;
     } else {
@@ -281,18 +344,22 @@ void MessagePack_Packer_module_init(VALUE mMessagePack)
     s_write = rb_intern("write");
 
     msgpack_packer_static_init();
+    msgpack_packer_ext_registry_static_init();
 
     cMessagePack_Packer = rb_define_class_under(mMessagePack, "Packer", rb_cObject);
 
-    rb_define_alloc_func(cMessagePack_Packer, Packer_alloc);
+    rb_define_alloc_func(cMessagePack_Packer, MessagePack_Packer_alloc);
 
-    rb_define_method(cMessagePack_Packer, "initialize", Packer_initialize, -1);
+    rb_define_method(cMessagePack_Packer, "initialize", MessagePack_Packer_initialize, -1);
+    rb_define_method(cMessagePack_Packer, "compatibility_mode?", Packer_compatibility_mode_p, 0);
     rb_define_method(cMessagePack_Packer, "buffer", Packer_buffer, 0);
     rb_define_method(cMessagePack_Packer, "write", Packer_write, 1);
     rb_define_alias(cMessagePack_Packer, "pack", "write");
     rb_define_method(cMessagePack_Packer, "write_nil", Packer_write_nil, 0);
     rb_define_method(cMessagePack_Packer, "write_array_header", Packer_write_array_header, 1);
     rb_define_method(cMessagePack_Packer, "write_map_header", Packer_write_map_header, 1);
+    rb_define_method(cMessagePack_Packer, "write_ext", Packer_write_ext, 2);
+    rb_define_method(cMessagePack_Packer, "write_float32", Packer_write_float32, 1);
     rb_define_method(cMessagePack_Packer, "flush", Packer_flush, 0);
 
     /* delegation methods */
@@ -306,7 +373,10 @@ void MessagePack_Packer_module_init(VALUE mMessagePack)
     //rb_define_method(cMessagePack_Packer, "append", Packer_append, 1);
     //rb_define_alias(cMessagePack_Packer, "<<", "append");
 
-    //s_packer_value = Packer_alloc(cMessagePack_Packer);
+    rb_define_private_method(cMessagePack_Packer, "registered_types_internal", Packer_registered_types_internal, 0);
+    rb_define_method(cMessagePack_Packer, "register_type", Packer_register_type, -1);
+
+    //s_packer_value = MessagePack_Packer_alloc(cMessagePack_Packer);
     //rb_gc_register_address(&s_packer_value);
     //Data_Get_Struct(s_packer_value, msgpack_packer_t, s_packer);
 
