@@ -19,6 +19,7 @@
 #include "unpacker.h"
 #include "unpacker_class.h"
 #include "buffer_class.h"
+#include "factory_class.h"
 
 VALUE cMessagePack_Unpacker;
 
@@ -28,7 +29,9 @@ VALUE cMessagePack_Unpacker;
 static VALUE eUnpackError;
 static VALUE eMalformedFormatError;
 static VALUE eStackError;
-static VALUE eTypeError;
+static VALUE eUnexpectedTypeError;
+static VALUE eUnknownExtTypeError;
+static VALUE mTypeError;  // obsoleted. only for backward compatibility. See #86.
 
 #define UNPACKER(from, name) \
     msgpack_unpacker_t *name = NULL; \
@@ -42,23 +45,31 @@ static void Unpacker_free(msgpack_unpacker_t* uk)
     if(uk == NULL) {
         return;
     }
+    msgpack_unpacker_ext_registry_destroy(&uk->ext_registry);
     _msgpack_unpacker_destroy(uk);
-    free(uk);
+    xfree(uk);
 }
 
-static VALUE Unpacker_alloc(VALUE klass)
+static void Unpacker_mark(msgpack_unpacker_t* uk)
+{
+    msgpack_unpacker_mark(uk);
+    msgpack_unpacker_ext_registry_mark(&uk->ext_registry);
+}
+
+VALUE MessagePack_Unpacker_alloc(VALUE klass)
 {
     msgpack_unpacker_t* uk = ALLOC_N(msgpack_unpacker_t, 1);
     _msgpack_unpacker_init(uk);
 
-    VALUE self = Data_Wrap_Struct(klass, msgpack_unpacker_mark, Unpacker_free, uk);
+    VALUE self = Data_Wrap_Struct(klass, Unpacker_mark, Unpacker_free, uk);
 
+    msgpack_unpacker_ext_registry_init(&uk->ext_registry);
     uk->buffer_ref = MessagePack_Buffer_wrap(UNPACKER_BUFFER_(uk), self);
 
     return self;
 }
 
-static VALUE Unpacker_initialize(int argc, VALUE* argv, VALUE self)
+VALUE MessagePack_Unpacker_initialize(int argc, VALUE* argv, VALUE self)
 {
     VALUE io = Qnil;
     VALUE options = Qnil;
@@ -90,21 +101,31 @@ static VALUE Unpacker_initialize(int argc, VALUE* argv, VALUE self)
 
     UNPACKER(self, uk);
 
-    MessagePack_Unpacker_initialize(uk, io, options);
-
-    return self;
-}
-
-void MessagePack_Unpacker_initialize(msgpack_unpacker_t* uk, VALUE io, VALUE options)
-{
-    MessagePack_Buffer_initialize(UNPACKER_BUFFER_(uk), io, options);
+    MessagePack_Buffer_set_options(UNPACKER_BUFFER_(uk), io, options);
 
     if(options != Qnil) {
         VALUE v;
 
         v = rb_hash_aref(options, ID2SYM(rb_intern("symbolize_keys")));
         msgpack_unpacker_set_symbolized_keys(uk, RTEST(v));
+
+        v = rb_hash_aref(options, ID2SYM(rb_intern("allow_unknown_ext")));
+        msgpack_unpacker_set_allow_unknown_ext(uk, RTEST(v));
     }
+
+    return self;
+}
+
+static VALUE Unpacker_symbolized_keys_p(VALUE self)
+{
+    UNPACKER(self, uk);
+    return uk->symbolize_keys ? Qtrue : Qfalse;
+}
+
+static VALUE Unpacker_allow_unknown_ext_p(VALUE self)
+{
+    UNPACKER(self, uk);
+    return uk->allow_unknown_ext ? Qtrue : Qfalse;
 }
 
 static void raise_unpacker_error(int r)
@@ -117,7 +138,9 @@ static void raise_unpacker_error(int r)
     case PRIMITIVE_STACK_TOO_DEEP:
         rb_raise(eStackError, "stack level too deep");
     case PRIMITIVE_UNEXPECTED_TYPE:
-        rb_raise(eTypeError, "unexpected type");
+        rb_raise(eUnexpectedTypeError, "unexpected type");
+    case PRIMITIVE_UNEXPECTED_EXT_TYPE:
+        rb_raise(eUnknownExtTypeError, "unexpected extension type");
     default:
         rb_raise(eUnpackError, "logically unknown error %d", r);
     }
@@ -283,6 +306,13 @@ static VALUE Unpacker_each(VALUE self)
 
 static VALUE Unpacker_feed_each(VALUE self, VALUE data)
 {
+#ifdef RETURN_ENUMERATOR
+    {
+        VALUE argv[] = { data };
+        RETURN_ENUMERATOR(self, sizeof(argv) / sizeof(VALUE), argv);
+    }
+#endif
+
     // TODO optimize
     Unpacker_feed(self, data);
     return Unpacker_each(self);
@@ -297,48 +327,93 @@ static VALUE Unpacker_reset(VALUE self)
     return Qnil;
 }
 
+static VALUE Unpacker_registered_types_internal(VALUE self)
+{
+    UNPACKER(self, uk);
+
+    VALUE mapping = rb_hash_new();
+    for(int i=0; i < 256; i++) {
+        if(uk->ext_registry.array[i] != Qnil) {
+            rb_hash_aset(mapping, INT2FIX(i - 128), uk->ext_registry.array[i]);
+        }
+    }
+
+    return mapping;
+}
+
+static VALUE Unpacker_register_type(int argc, VALUE* argv, VALUE self)
+{
+    UNPACKER(self, uk);
+
+    int ext_type;
+    VALUE proc;
+    VALUE arg;
+    VALUE ext_class;
+
+    switch (argc) {
+    case 1:
+        /* register_type(0x7f) {|data| block... } */
+        rb_need_block();
+#ifdef HAVE_RB_BLOCK_LAMBDA
+        proc = rb_block_lambda();
+#else
+        /* MRI 1.8 */
+        proc = rb_block_proc();
+#endif
+        arg = proc;
+        ext_class = Qnil;
+        break;
+    case 3:
+        /* register_type(0x7f, Time, :from_msgpack_ext) */
+        ext_class = argv[1];
+        arg = argv[2];
+        proc = rb_obj_method(ext_class, arg);
+        break;
+    default:
+        rb_raise(rb_eArgError, "wrong number of arguments (%d for 1 or 3)", argc);
+    }
+
+    ext_type = NUM2INT(argv[0]);
+    if(ext_type < -128 || ext_type > 127) {
+        rb_raise(rb_eRangeError, "integer %d too big to convert to `signed char'", ext_type);
+    }
+
+    msgpack_unpacker_ext_registry_put(&uk->ext_registry, ext_class, ext_type, proc, arg);
+
+    return Qnil;
+}
+
 VALUE MessagePack_unpack(int argc, VALUE* argv)
 {
     VALUE src;
-    VALUE options = Qnil;
+    VALUE self;
 
-    switch(argc) {
-    case 2:
-        options = argv[1];
-        if(rb_type(options) != T_HASH) {
-            rb_raise(rb_eArgError, "expected Hash but found %s.", rb_obj_classname(options));
-        }
-        /* pass-through */
-    case 1:
-        src = argv[0];
-        break;
-    default:
+    if (argc < 0 || argc > 2) {
         rb_raise(rb_eArgError, "wrong number of arguments (%d for 1..2)", argc);
     }
+    src = argv[0];
 
-    VALUE self = Unpacker_alloc(cMessagePack_Unpacker);
+    if(rb_type(src) == T_STRING) {
+        self = MessagePack_Factory_unpacker(argc - 1, argv + 1, cMessagePack_DefaultFactory);
+        UNPACKER(self, uk);
+        msgpack_buffer_append_string(UNPACKER_BUFFER_(uk), src);
+    } else {
+        self = MessagePack_Factory_unpacker(argc, argv, cMessagePack_DefaultFactory);
+    }
     UNPACKER(self, uk);
-    //_msgpack_unpacker_reset(s_unpacker);
-    //msgpack_buffer_reset_io(UNPACKER_BUFFER_(s_unpacker));
 
     /* prefer reference than copying; see MessagePack_Unpacker_module_init */
     msgpack_buffer_set_write_reference_threshold(UNPACKER_BUFFER_(uk), 0);
 
-    if(rb_type(src) == T_STRING) {
-        MessagePack_Unpacker_initialize(uk, Qnil, options);
-        msgpack_buffer_append_string(UNPACKER_BUFFER_(uk), src);
-    } else {
-        MessagePack_Unpacker_initialize(uk, src, options);
-    }
-
     int r = msgpack_unpacker_read(uk, 0);
     if(r < 0) {
         raise_unpacker_error(r);
     }
 
     /* raise if extra bytes follow */
-    if(msgpack_buffer_top_readable_size(UNPACKER_BUFFER_(uk)) > 0) {
-        rb_raise(eMalformedFormatError, "extra bytes follow after a deserialized object");
+    size_t extra = msgpack_buffer_top_readable_size(UNPACKER_BUFFER_(uk));
+    if(extra > 0) {
+        rb_raise(eMalformedFormatError, "%zd extra bytes after the deserialized object", extra);
     }
 
 #ifdef RB_GC_GUARD
@@ -362,9 +437,19 @@ static VALUE MessagePack_unpack_module_method(int argc, VALUE* argv, VALUE mod)
     return MessagePack_unpack(argc, argv);
 }
 
+VALUE MessagePack_Unpacker_new(int argc, VALUE* argv)
+{
+    VALUE self = MessagePack_Unpacker_alloc(cMessagePack_Unpacker);
+    MessagePack_Unpacker_initialize(argc, argv, self);
+    return self;
+}
+
 void MessagePack_Unpacker_module_init(VALUE mMessagePack)
 {
     msgpack_unpacker_static_init();
+    msgpack_unpacker_ext_registry_static_init();
+
+    mTypeError = rb_define_module_under(mMessagePack, "TypeError");
 
     cMessagePack_Unpacker = rb_define_class_under(mMessagePack, "Unpacker", rb_cObject);
 
@@ -374,11 +459,16 @@ void MessagePack_Unpacker_module_init(VALUE mMessagePack)
 
     eStackError = rb_define_class_under(mMessagePack, "StackError", eUnpackError);
 
-    eTypeError = rb_define_class_under(mMessagePack, "TypeError", rb_eStandardError);
+    eUnexpectedTypeError = rb_define_class_under(mMessagePack, "UnexpectedTypeError", eUnpackError);
+    rb_include_module(eUnexpectedTypeError, mTypeError);
 
-    rb_define_alloc_func(cMessagePack_Unpacker, Unpacker_alloc);
+    eUnknownExtTypeError = rb_define_class_under(mMessagePack, "UnknownExtTypeError", eUnpackError);
 
-    rb_define_method(cMessagePack_Unpacker, "initialize", Unpacker_initialize, -1);
+    rb_define_alloc_func(cMessagePack_Unpacker, MessagePack_Unpacker_alloc);
+
+    rb_define_method(cMessagePack_Unpacker, "initialize", MessagePack_Unpacker_initialize, -1);
+    rb_define_method(cMessagePack_Unpacker, "symbolize_keys?", Unpacker_symbolized_keys_p, 0);
+    rb_define_method(cMessagePack_Unpacker, "allow_unknown_ext?", Unpacker_allow_unknown_ext_p, 0);
     rb_define_method(cMessagePack_Unpacker, "buffer", Unpacker_buffer, 0);
     rb_define_method(cMessagePack_Unpacker, "read", Unpacker_read, 0);
     rb_define_alias(cMessagePack_Unpacker, "unpack", "read");
@@ -392,7 +482,10 @@ void MessagePack_Unpacker_module_init(VALUE mMessagePack)
     rb_define_method(cMessagePack_Unpacker, "feed_each", Unpacker_feed_each, 1);
     rb_define_method(cMessagePack_Unpacker, "reset", Unpacker_reset, 0);
 
-    //s_unpacker_value = Unpacker_alloc(cMessagePack_Unpacker);
+    rb_define_private_method(cMessagePack_Unpacker, "registered_types_internal", Unpacker_registered_types_internal, 0);
+    rb_define_method(cMessagePack_Unpacker, "register_type", Unpacker_register_type, -1);
+
+    //s_unpacker_value = MessagePack_Unpacker_alloc(cMessagePack_Unpacker);
     //rb_gc_register_address(&s_unpacker_value);
     //Data_Get_Struct(s_unpacker_value, msgpack_unpacker_t, s_unpacker);
     /* prefer reference than copying */
